@@ -15,9 +15,9 @@ import time
 import itk
 import numpy as np
 import ipywidgets as widgets
-from traitlets import CBool, CFloat, Unicode, CaselessStrEnum, List, validate
+from traitlets import CBool, CFloat, CInt, Unicode, CaselessStrEnum, List, validate, TraitError, Tuple
 from ipydatawidgets import NDArray, array_serialization, shape_constraints
-from .trait_types import ITKImage, PointSetList, PolyDataList, itkimage_serialization, polydata_list_serialization, Colormap
+from .trait_types import ITKImage, ImagePointTrait, ImagePoint, PointSetList, PolyDataList, itkimage_serialization, image_point_serialization, polydata_list_serialization, Colormap
 
 try:
     import ipywebrtc
@@ -116,8 +116,8 @@ class Viewer(ViewerParent):
     _model_name = Unicode('ViewerModel').tag(sync=True)
     _view_module = Unicode('itkwidgets').tag(sync=True)
     _model_module = Unicode('itkwidgets').tag(sync=True)
-    _view_module_version = Unicode('^0.27.5').tag(sync=True)
-    _model_module_version = Unicode('^0.27.5').tag(sync=True)
+    _view_module_version = Unicode('^0.28.0').tag(sync=True)
+    _model_module_version = Unicode('^0.28.0').tag(sync=True)
     image = ITKImage(
         default_value=None,
         allow_none=True,
@@ -132,6 +132,27 @@ class Viewer(ViewerParent):
     _rendering_image = CBool(
         default_value=False,
         help="We are currently volume rendering the image.").tag(sync=True)
+    label_map = ITKImage(
+        default_value=None,
+        allow_none=True,
+        help="Label map for the image.").tag(
+        sync=False,
+        **itkimage_serialization)
+    rendered_label_map = ITKImage(
+        default_value=None,
+        allow_none=True).tag(
+        sync=True,
+        **itkimage_serialization)
+    label_map_names = List(
+        trait=Tuple(),
+        allow_none=True,
+        default_value=None,
+        help="Names for labels in the label map.").tag(
+        sync=True)
+    label_map_weights = NDArray(dtype=np.float32, default_value=None, allow_none=True,
+        help="Weights, from 0.0 to 1.0, for every label in the label map.")\
+        .tag(sync=True, **array_serialization)\
+        .valid(shape_constraints(None,))
     interpolation = CBool(
         default_value=True,
         help="Use linear interpolation in slicing planes.").tag(sync=True)
@@ -159,6 +180,12 @@ class Viewer(ViewerParent):
         default_value=None,
         allow_none=True,
         help="World-space position of the Z slicing plane.").tag(sync=True)
+    clicked_slice_point = ImagePointTrait(
+        default_value=None,
+        allow_none=True,
+        help="Data for the point clicked on an image slice.").tag(
+            sync=True,
+            **image_point_serialization)
     gradient_opacity = CFloat(
         default_value=0.2,
         help="Volume rendering gradient opacity, from (0.0, 1.0]").tag(sync=True)
@@ -295,26 +322,40 @@ class Viewer(ViewerParent):
             opacities_array = self._validate_geometry_opacities(proposal)
             kwargs['geometry_opacities'] = opacities_array
         self.observe(self._on_geometries_changed, ['geometries'])
+        have_label_map = 'label_map' in kwargs and kwargs['label_map'] is not None
+        if have_label_map:
+            # Interpolation is not currently supported with label maps
+            kwargs['interpolation'] = False
+        if not 'cmap' in kwargs or kwargs['cmap'] is None:
+            if have_label_map:
+                kwargs['cmap'] = cm.grayscale
+            else:
+                kwargs['cmap'] = cm.viridis
 
         super(Viewer, self).__init__(**kwargs)
 
-        if not self.image:
+        if not self.image and not self.label_map:
             return
-        dimension = self.image.GetImageDimension()
-        largest_region = self.image.GetLargestPossibleRegion()
+        if self.image:
+            image = self.image
+        else:
+            image = self.label_map
+        dimension = image.GetImageDimension()
+        largest_region = image.GetLargestPossibleRegion()
         size = largest_region.GetSize()
 
         # Cache this so we do not need to recompute on it when resetting the
         # roi
         self._largest_roi_rendered_image = None
+        self._largest_roi_rendered_label_map = None
         self._largest_roi = np.zeros((2, 3), dtype=np.float64)
         if not np.any(self.roi):
             largest_index = largest_region.GetIndex()
             self.roi[0][:dimension] = np.array(
-                self.image.TransformIndexToPhysicalPoint(largest_index))
+                image.TransformIndexToPhysicalPoint(largest_index))
             largest_index_upper = largest_index + size
             self.roi[1][:dimension] = np.array(
-                self.image.TransformIndexToPhysicalPoint(largest_index_upper))
+                image.TransformIndexToPhysicalPoint(largest_index_upper))
             self._largest_roi = self.roi.copy()
 
         if dimension == 2:
@@ -325,15 +366,19 @@ class Viewer(ViewerParent):
             for dim in range(dimension):
                 if size[dim] > self.size_limit_3d[dim]:
                     self._downsampling = True
-        if self._downsampling:
+        if self._downsampling and self.image:
             self.extractor = itk.ExtractImageFilter.New(self.image)
             self.shrinker = itk.BinShrinkImageFilter.New(self.extractor)
+        if self._downsampling and self.label_map:
+            self.label_map_extractor = itk.ExtractImageFilter.New(self.label_map)
+            self.label_map_shrinker = itk.ShrinkImageFilter.New(self.label_map_extractor)
         self._update_rendered_image()
         if self._downsampling:
             self.observe(self._on_roi_changed, ['roi'])
 
         self.observe(self._on_reset_crop_requested, ['_reset_crop_requested'])
-        self.observe(self.update_rendered_image, ['image'])
+        self.observe(self.update_rendered_image, ['image', 'label_map'])
+        self.observe(self.update_rendered_image, ['image', 'label_map'])
 
     def _on_roi_changed(self, change=None):
         if self._downsampling:
@@ -341,16 +386,20 @@ class Viewer(ViewerParent):
 
     def _on_reset_crop_requested(self, change=None):
         if change.new is True and self._downsampling:
-            dimension = self.image.GetImageDimension()
-            largest_region = self.image.GetLargestPossibleRegion()
+            if self.image:
+                image = self.image
+            else:
+                image = self.label_map
+            dimension = image.GetImageDimension()
+            largest_region = image.GetLargestPossibleRegion()
             size = largest_region.GetSize()
             largest_index = largest_region.GetIndex()
             new_roi = self.roi.copy()
             new_roi[0][:dimension] = np.array(
-                self.image.TransformIndexToPhysicalPoint(largest_index))
+                image.TransformIndexToPhysicalPoint(largest_index))
             largest_index_upper = largest_index + size
             new_roi[1][:dimension] = np.array(
-                self.image.TransformIndexToPhysicalPoint(largest_index_upper))
+                image.TransformIndexToPhysicalPoint(largest_index_upper))
             self._largest_roi = new_roi.copy()
             self.roi = new_roi
         if change.new is True:
@@ -359,6 +408,7 @@ class Viewer(ViewerParent):
     @debounced(delay_seconds=0.2, method=True)
     def update_rendered_image(self, change=None):
         self._largest_roi_rendered_image = None
+        self._largest_roi_rendered_label_map = None
         self._largest_roi = np.zeros((2, 3), dtype=np.float64)
         self._update_rendered_image()
 
@@ -371,7 +421,7 @@ class Viewer(ViewerParent):
         return scale_factors
 
     def _update_rendered_image(self):
-        if self.image is None:
+        if self.image is None and self.label_map is None:
             return
         if self._rendering_image:
             @yield_for_change(self, '_rendering_image')
@@ -382,10 +432,14 @@ class Viewer(ViewerParent):
         self._rendering_image = True
 
         if self._downsampling:
-            dimension = self.image.GetImageDimension()
-            index = self.image.TransformPhysicalPointToIndex(
+            if self.image:
+                image = self.image
+            else:
+                image = self.label_map
+            dimension = image.GetImageDimension()
+            index = image.TransformPhysicalPointToIndex(
                 self.roi[0][:dimension])
-            upper_index = self.image.TransformPhysicalPointToIndex(
+            upper_index = image.TransformPhysicalPointToIndex(
                 self.roi[1][:dimension])
             size = upper_index - index
 
@@ -396,7 +450,10 @@ class Viewer(ViewerParent):
                 scale_factors = self._find_scale_factors(
                     self.size_limit_3d, dimension, size)
             self._scale_factors = np.array(scale_factors, dtype=np.uint8)
-            self.shrinker.SetShrinkFactors(scale_factors[:dimension])
+            if self.image:
+                self.shrinker.SetShrinkFactors(scale_factors[:dimension])
+            if self.label_map:
+                self.label_map_shrinker.SetShrinkFactors(scale_factors[:dimension])
 
             region = itk.ImageRegion[dimension]()
             region.SetIndex(index)
@@ -404,10 +461,14 @@ class Viewer(ViewerParent):
             # Account for rounding
             # truncation issues
             region.PadByRadius(1)
-            region.Crop(self.image.GetLargestPossibleRegion())
+            region.Crop(image.GetLargestPossibleRegion())
 
-            self.extractor.SetInput(self.image)
-            self.extractor.SetExtractionRegion(region)
+            if self.image:
+                self.extractor.SetInput(self.image)
+                self.extractor.SetExtractionRegion(region)
+            if self.label_map:
+                self.label_map_extractor.SetInput(self.label_map)
+                self.label_map_extractor.SetExtractionRegion(region)
 
             size = region.GetSize()
 
@@ -415,24 +476,57 @@ class Viewer(ViewerParent):
             if np.any(self._largest_roi) and np.all(
                     self._largest_roi == self.roi):
                 is_largest = True
-                if self._largest_roi_rendered_image is not None:
-                    self.rendered_image = self._largest_roi_rendered_image
+                if self._largest_roi_rendered_image is not None or self._largest_roi_rendered_label_map is not None:
+                    if self.image:
+                        self.rendered_image = self._largest_roi_rendered_image
+                    if self.label_map:
+                        self.rendered_label_map = self._largest_roi_rendered_label_map
                     return
 
-            self.shrinker.UpdateLargestPossibleRegion()
+            if self.image:
+                self.shrinker.UpdateLargestPossibleRegion()
+            if self.label_map:
+                self.label_map_shrinker.UpdateLargestPossibleRegion()
             if is_largest:
-                self._largest_roi_rendered_image = self.shrinker.GetOutput()
-                self._largest_roi_rendered_image.DisconnectPipeline()
-                self._largest_roi_rendered_image.SetOrigin(
-                    self.roi[0][:dimension])
-                self.rendered_image = self._largest_roi_rendered_image
+                if self.image:
+                    self._largest_roi_rendered_image = self.shrinker.GetOutput()
+                    self._largest_roi_rendered_image.DisconnectPipeline()
+                    self._largest_roi_rendered_image.SetOrigin(
+                        self.roi[0][:dimension])
+                    self.rendered_image = self._largest_roi_rendered_image
+                if self.label_map:
+                    self._largest_roi_rendered_label_map = self.label_map_shrinker.GetOutput()
+                    self._largest_roi_rendered_label_map.DisconnectPipeline()
+                    self._largest_roi_rendered_label_map.SetOrigin(
+                        self.roi[0][:dimension])
+                    self.rendered_label_map = self._largest_roi_rendered_label_map
                 return
-            shrunk = self.shrinker.GetOutput()
-            shrunk.DisconnectPipeline()
-            shrunk.SetOrigin(self.roi[0][:dimension])
-            self.rendered_image = shrunk
+            if self.image:
+                shrunk = self.shrinker.GetOutput()
+                shrunk.DisconnectPipeline()
+                shrunk.SetOrigin(self.roi[0][:dimension])
+                self.rendered_image = shrunk
+            if self.label_map:
+                shrunk = self.label_map_shrinker.GetOutput()
+                shrunk.DisconnectPipeline()
+                shrunk.SetOrigin(self.roi[0][:dimension])
+                self.rendered_label_map = shrunk
         else:
-            self.rendered_image = self.image
+            if self.image:
+                self.rendered_image = self.image
+            if self.label_map:
+                self.rendered_label_map = self.image
+
+    @validate('label_map_weights')
+    def _validate_label_map_weights(self, proposal):
+        """Check the number of weights equals the number of labels."""
+        value = proposal['value']
+        value = np.array(value, dtype=np.float32)
+        if self.rendered_label_map:
+            labels = len(np.unique(itk.array_view_from_image(self.rendered_label_map)))
+            if labels != len(value):
+                raise TraitError('Number of labels, {0}, does not equal number of label weights, {1}'.format(labels, len(value)))
+        return value
 
     @validate('gradient_opacity')
     def _validate_gradient_opacity(self, proposal):
@@ -541,10 +635,14 @@ class Viewer(ViewerParent):
 
     def roi_region(self):
         """Return the itk.ImageRegion corresponding to the roi."""
-        dimension = self.image.GetImageDimension()
-        index = self.image.TransformPhysicalPointToIndex(
+        if self.image:
+            image = self.image
+        else:
+            image = self.label_map
+        dimension = image.GetImageDimension()
+        index = image.TransformPhysicalPointToIndex(
             tuple(self.roi[0][:dimension]))
-        upper_index = self.image.TransformPhysicalPointToIndex(
+        upper_index = image.TransformPhysicalPointToIndex(
             tuple(self.roi[1][:dimension]))
         size = upper_index - index
         for dim in range(dimension):
@@ -552,12 +650,16 @@ class Viewer(ViewerParent):
         region = itk.ImageRegion[dimension]()
         region.SetIndex(index)
         region.SetSize(tuple(size))
-        region.Crop(self.image.GetLargestPossibleRegion())
+        region.Crop(image.GetLargestPossibleRegion())
         return region
 
     def roi_slice(self):
         """Return the numpy array slice corresponding to the roi."""
-        dimension = self.image.GetImageDimension()
+        if self.image:
+            image = self.image
+        else:
+            image = self.label_map
+        dimension = image.GetImageDimension()
         region = self.roi_region()
         index = region.GetIndex()
         upper_index = np.array(index) + np.array(region.GetSize())
@@ -568,7 +670,10 @@ class Viewer(ViewerParent):
 
 
 def view(image=None,  # noqa: C901
-         cmap=cm.viridis,
+         label_map=None,  # noqa: C901
+         label_map_names=None,  # noqa: C901
+         label_map_weights=None,  # noqa: C901
+         cmap=None,
          select_roi=False,
          interpolation=True,
          gradient_opacity=0.22, slicing_planes=False, shadow=True, blend='composite',
@@ -584,7 +689,8 @@ def view(image=None,  # noqa: C901
     Creates and returns an ipywidget to visualize an image, and/or point sets
     and/or geometries .
 
-    The image can be 2D or 3D.
+    The image can be 2D or 3D. A label map that corresponds to the image can
+    also be provided. The image and label map must have the same size.
 
     The type of the image can be an numpy.array, itk.Image,
     vtk.vtkImageData, pyvista.UniformGrid, imglyb.ReferenceGuardingRandomAccessibleInterval,
@@ -602,18 +708,18 @@ def view(image=None,  # noqa: C901
     General Interface
     ^^^^^^^^^^^^^^^^^
 
-    ui_collapsed : bool, optional, default: False
+    ui_collapsed : bool, default: False
         Collapse the native widget user interface.
 
-    rotate : bool, optional, default: False
+    rotate : bool, default: False
         Continuously rotate the camera around the scene in volume rendering
         mode.
 
-    annotations : bool, optional, default: True
+    annotations : bool, default: True
         Display annotations describing orientation and the value of a
         mouse-position-based data probe.
 
-    mode: 'x', 'y', 'z', or 'v', optional, default: 'v'
+    mode: 'x', 'y', 'z', or 'v', default: 'v'
         Only relevant for 3D scenes.
         Viewing mode:
             'x': x-plane
@@ -621,7 +727,7 @@ def view(image=None,  # noqa: C901
             'z': z-plane
             'v': volume rendering
 
-    camera: 3x3 numpy float32 array, optional
+    camera: 3x3 numpy float32 array
         Camera parameters:
             [[position_x,    position_y,    position_z],
              [focal_point_x, focal_point_y, focal_point_z],
@@ -634,88 +740,99 @@ def view(image=None,  # noqa: C901
     image : array_like, itk.Image, or vtk.vtkImageData
         The 2D or 3D image to visualize.
 
-    vmin: float, optional, default: None
+    label_map : array_like, itk.Image, or vtk.vtkImageData
+        The 2D or 3D label map to visualize. If an image is also provided, the
+        label map must have the same size.
+
+    label_map_names : OrderedDict of (label_value, label_name)
+        String names associated with the integer label values.
+
+    label_map_weights : 1D numpy float32 array, default: None
+        Rendering weights, from 0.0 to 1.0, associated labels in the label map.
+
+    vmin: float, default: None
         Value that maps to the minimum of image colormap. Defaults to minimum of
         the image pixel buffer.
 
-    vmax: float, optional, default: None
+    vmax: float, default: None
         Value that maps to the minimum of image colormap. Defaults to maximum of
         the image pixel buffer.
 
-    cmap: string, optional, default: 'Viridis (matplotlib)'
+    cmap: string, default: viridis, grayscale with a label map
         Colormap. Some valid values available at itkwidgets.cm.*
 
-    select_roi: bool, optional, default: False
+    select_roi: bool, default: False
         Enable an interactive region of interest widget for the image.
 
-    slicing_planes: bool, optional, default: False
+    slicing_planes: bool, default: False
         Enable slicing planes on the volume rendering.
 
-    x_slice: float, optional, default: None
+    x_slice: float, default: None
         World-space position of the X slicing plane.
 
-    y_slice: float, optional, default: None
+    y_slice: float, default: None
         World-space position of the Y slicing plane.
 
-    z_slice: float, optional, default: None
+    z_slice: float, default: None
         World-space position of the Z slicing plane.
 
-    interpolation: bool, optional, default: True
+    interpolation: bool, default: True
         Linear as opposed to nearest neighbor interpolation for image slices.
+        Note: Interpolation is not currently supported with label maps.
 
-    gradient_opacity: float, optional, default: 0.22
+    gradient_opacity: float, default: 0.22
         Gradient opacity for composite volume rendering, in the range (0.0, 1.0].
 
-    shadow: bool, optional, default: True
+    shadow: bool, default: True
         Use shadowing with composite volume rendering.
 
-    blend: 'composite', 'max', 'min', or 'average', optional, default: 'composite'
+    blend: 'composite', 'max', 'min', or 'average', default: 'composite'
         Volume rendering blend mode.
 
     Point Sets
     ^^^^^^^^^^
 
-    point_sets: point set, or sequence of point sets, optional
+    point_sets: point set, or sequence of point sets
         The point sets to visualize.
 
-    point_set_colors: list of RGB colors, optional
+    point_set_colors: list of RGB colors
         Colors for the N geometries. See help(matplotlib.colors) for
         specification. Defaults to the Glasbey series of categorical colors.
 
-    point_set_opacities: list of floats, optional, default: [0.5,]*n
+    point_set_opacities: list of floats, default: [0.5,]*n
         Opacity for the point sets, in the range (0.0, 1.0].
 
-    point_set_representations: list of strings, optional, default: ['points',]*n
+    point_set_representations: list of strings, default: ['points',]*n
         How to represent the point set. One of 'hidden', 'points', or 'spheres'.
 
     Geometries
     ^^^^^^^^^^
 
-    geometries: geometries, or sequence of geometries, optional
+    geometries: geometries, or sequence of geometries
         The geometries to visualize.
 
-    geometry_colors: list of RGB colors, optional
+    geometry_colors: list of RGB colors
         Colors for the N geometries. See help(matplotlib.colors) for
         specification. Defaults to the Glasbey series of categorical colors.
 
-    geometry_opacities: list of floats, optional, default: [1.0,]*n
+    geometry_opacities: list of floats, default: [1.0,]*n
         Opacity for the point sets, in the range (0.0, 1.0].
 
 
     Other Parameters
     ----------------
 
-    units: string, optional, default: ''
+    units: string, default: ''
         Units to display in the scale bar.
 
-    actors: vtkActor, vtkAssembly, vtkVolume, optional, default: None
+    actors: vtkActor, vtkAssembly, vtkVolume, default: None
         List of standard vtk objects, colors are extracted from their properties
 
-    size_limit_2d: 2x1 numpy int64 array, optional, default: [1024, 1024]
+    size_limit_2d: 2x1 numpy int64 array, default: [1024, 1024]
         Size limit for 2D image visualization. If the roi is larger than this
         size, it will be downsampled for visualization
 
-    size_limit_3d: 3x1 numpy int64 array, optional, default: [192, 192, 192]
+    size_limit_3d: 3x1 numpy int64 array, default: [192, 192, 192]
         Size limit for 3D image visualization. If the roi is larger than this
         size, it will be downsampled for visualization.
 
@@ -798,6 +915,9 @@ def view(image=None,  # noqa: C901
             image = images[0]
 
     viewer = Viewer(image=image,
+                    label_map=label_map,
+                    label_map_names=label_map_names,
+                    label_map_weights=label_map_weights,
                     cmap=cmap,
                     select_roi=select_roi,
                     interpolation=interpolation,
